'----------------------------------------------------------------------------------------
'    Name:  Modbus_CalcCRC_i.inc
'    Type:  TIGER-BASIC(tm) Source Code
' Purpose:  Subroutines for Modbus checksum calculation
'
'  (C) - Copyright Wilke Technology, P.O.Box 1727, D-52018 Aachen, Germany
'----------------------------------------------------------------------------------------
'
' Thank you for using BASIC Tigers in your products. If you have questions, ideas
' or special needs, please contact your next distributor or the Tiger support team
' and visit our web site:
'
'	Wilke Technology GmbH
'	The Tiger Support Team
'	P.O.Box 1727,       D-52018 Aachen, Germany
'	Krefelder Str. 147, D-52070 Aachen, Germany
'
'	email: support@wilke-technology.com  (english)
'	email: support@wilke.de              (german)
'	Phone: +49 (241) 918 900    Mo to Fr, 7:00 to 16:00 (GMT)
'	Fax:   +49 (241) 918 9068
'
' New information, new drivers and free downloads see:
'
'	www.wilke-technology.com	     (english)
'	www.wilke.de                         (german)
'
' Sincerely,
'
'	Your Tiger Support Team
'
'
'--------------------------------------------------------------------------------------

#ifndef Modbus_CalcCRC_i
#define Modbus_CalcCRC_i
#include Modbus_CalcCRC_d.inc

'------------------------------------------------------------------------------------------------
'******* Use for CRC-calculation ********************
ARRAY abgCRCHi(256) OF BYTE
ARRAY abgCRCLo(256) OF BYTE
DATALABEL dagCRChi,dagCRClo
'****************************************************
''***************************************************
''***	Calculate CRC
''***************************************************
sub calcCRC(string spDataSend$; var word wvpCRC)
  byte blCRCHi, blCRCLo, blChar, blIndex
  long llDataLen
  string sltemp$
  word wlIndex

  blCRCHi = HIGH_CRC_BYTE 															' high CRC byte initialized
  blCRCLo = LOW_CRC_BYTE 															' low CRC byte initialized

  llDataLen=len(spDataSend$)
  for wlIndex=0 to (llDataLen-1)
    sltemp$=mid$(spDataSend$, wlIndex, SZE_OF_BYTE)
    blchar=nfroms(sltemp$, 0, SZE_OF_BYTE)
    blIndex = blCRCHi bitxor blChar
    blCRCHi = blCRCLo bitxor abgCRCHi(blIndex)
    blCRCLo = abgCRCLo(blIndex)
  next
  wvpCRC=(blCRCHi SHL BIT_NUM_OF_BYTE) bitor blCRCLo
end

''***************************************************
''***	Calculate LRC
''***************************************************
sub calcLRC(byte bpDataLen ;string spDataSend$; var byte bvpLRCval)
  byte 	blCurPos, blLength, blLRC
  string	Temp$
  byte blChar

  blLRC=0
  blLength=bpDataLen
  blCurPos=0
  while blLength>0
    Temp$=mid$(spDataSend$, blCurPos, SZE_OF_BYTE)
    blChar=nfroms(temp$,0,SZE_OF_BYTE)
    blLRC=blLRC+blChar
    blCurPos=blCurPos+1
    blLength=blLength-1
  endwhile
  bvpLRCval=(blLRC bitxor HIGH_BYTE_VALUE)+1
end

''***************************************************
''***	Init. Datalabel for CRC
''***************************************************

sub CRCinit
  byte blIndex
  ' High Order Byte Table > Table of CRC values for high-order byte
  dagCRChi::
  data byte &
  000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, &
  040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, &
  080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, &
  0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, &
  000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, &
  040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, &
  080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, &
  0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, &
  000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, &
  040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, &
  080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, &
  0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, &
  000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, &
  040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, &
  080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, &
  0C0h, 080h, 041h, 000h, 0C1h, 081h, 040h, 000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, &
  000h, 0C1h, 081h, 040h, 001h, 0C0h, 080h, 041h, 001h, 0C0h, 080h, 041h, 000h, 0C1h, 081h, &
  040h

  ' Low Order Byte Table > Table of CRC values for low-order byte
  dagCRClo::
  data byte &
  000h, 0C0h, 0C1h, 001h, 0C3h, 003h, 002h, 0C2h, 0C6h, 006h, 007h, 0C7h, 005h, 0C5h, 0C4h, &
  004h, 0CCh, 00Ch, 00Dh, 0CDh, 00Fh, 0CFh, 0CEh, 00Eh, 00Ah, 0CAh, 0CBh, 00Bh, 0C9h, 009h, &
  008h, 0C8h, 0D8h, 018h, 019h, 0D9h, 01Bh, 0DBh, 0DAh, 01Ah, 01Eh, 0DEh, 0DFh, 01Fh, 0DDh, &
  01Dh, 01Ch, 0DCh, 014h, 0D4h, 0D5h, 015h, 0D7h, 017h, 016h, 0D6h, 0D2h, 012h, 013h, 0D3h, &
  011h, 0D1h, 0D0h, 010h, 0F0h, 030h, 031h, 0F1h, 033h, 0F3h, 0F2h, 032h, 036h, 0F6h, 0F7h, &
  037h, 0F5h, 035h, 034h, 0F4h, 03Ch, 0FCh, 0FDh, 03Dh, 0FFh, 03Fh, 03Eh, 0FEh, 0FAh, 03Ah, &
  03Bh, 0FBh, 039h, 0F9h, 0F8h, 038h, 028h, 0E8h, 0E9h, 029h, 0EBh, 02Bh, 02Ah, 0EAh, 0EEh, &
  02Eh, 02Fh, 0EFh, 02Dh, 0EDh, 0ECh, 02Ch, 0E4h, 024h, 025h, 0E5h, 027h, 0E7h, 0E6h, 026h, &
  022h, 0E2h, 0E3h, 023h, 0E1h, 021h, 020h, 0E0h, 0A0h, 060h, 061h, 0A1h, 063h, 0A3h, 0A2h, &
  062h, 066h, 0A6h, 0A7h, 067h, 0A5h, 065h, 064h, 0A4h, 06Ch, 0ACh, 0ADh, 06Dh, 0AFh, 06Fh, &
  06Eh, 0AEh, 0AAh, 06Ah, 06Bh, 0ABh, 069h, 0A9h, 0A8h, 068h, 078h, 0B8h, 0B9h, 079h, 0BBh, &
  07Bh, 07Ah, 0BAh, 0BEh, 07Eh, 07Fh, 0BFh, 07Dh, 0BDh, 0BCh, 07Ch, 0B4h, 074h, 075h, 0B5h, &
  077h, 0B7h, 0B6h, 076h, 072h, 0B2h, 0B3h, 073h, 0B1h, 071h, 070h, 0B0h, 050h, 090h, 091h, &
  051h, 093h, 053h, 052h, 092h, 096h, 056h, 057h, 097h, 055h, 095h, 094h, 054h, 09Ch, 05Ch, &
  05Dh, 09Dh, 05Fh, 09Fh, 09Eh, 05Eh, 05Ah, 09Ah, 09Bh, 05Bh, 099h, 059h, 058h, 098h, 088h, &
  048h, 049h, 089h, 04Bh, 08Bh, 08Ah, 04Ah, 04Eh, 08Eh, 08Fh, 04Fh, 08Dh, 04Dh, 04Ch, 08Ch, &
  044h, 084h, 085h, 045h, 087h, 047h, 046h, 086h, 082h, 042h, 043h, 083h, 041h, 081h, 080h, &
  040h


  '' Read from flash each with length 1
  for blIndex=0 to LEN_OF_CRC_DATALABEL
    Peek_Flash dagCRClo+blIndex,abgCRCLo(blIndex),1
    Peek_Flash dagCRChi+blIndex,abgCRCHi(blIndex),1
  next
end
''
'------------------------------------------------------------------------------------------------
#endif ''Modbus_CalcCRC_i

