'----------------------------------------------------------------------------------------
'    Name:  Modbus_Slave_Functions.inc
'    Type:  TIGER-BASIC(tm) Source Code (Modbus on RS485-bus)
' Release:  V 1.0
' Purpose:  ModbusProtocol (Slave) on RS485
'           (for information about the Modbus protocol please visit www.modbus.org)
'
'  (C) - Copyright Wilke Technology, P.O.Box 1727, D-52018 Aachen, Germany
'----------------------------------------------------------------------------------------
'
' Thank you for using BASIC Tigers in your products. If you have questions, ideas
' or special needs, please contact your next distributor or the Tiger support team
' and visit our web site:
'
'	Wilke Technology GmbH
'	The Tiger Support Team
'	P.O.Box 1727,       D-52018 Aachen, Germany
'	Krefelder Str. 147, D-52070 Aachen, Germany
'
'	email: support@wilke-technology.com  (english)
'	email: support@wilke.de              (german)
'	Phone: +49 (241) 918 900    Mo to Fr, 7:00 to 16:00 (GMT)
'	Fax:   +49 (241) 918 9068
'
' New information, new drivers and free downloads see:
'
'	www.wilke-technology.com	     (english)
'	www.wilke.de                         (german)
'
' Sincerely,
'
'	Your Tiger Support Team
'
'
'--------------------------------------------------------------------------------------
' Supported Mobbus-Slave-Functions:
' 0x01 - Read Coils                         ' SReadCoils
' 0x02 - Read Discrete Inputs               ' SReadDisInputs
' 0x03 - Read Hold Reg                      ' SReadHoldReg
' 0x04 - Read Input Reg                     ' SReadInputReg
' 0x05 - Write Single Coil                  ' SWriteSingleCoil
' 0x06 - Write Single Reg                   ' SWriteSingleReg
' 0x07 - Read Exception Reg                 ' SReadExceptionReg
' 0x0B - Get Com Event Counter              ' SGetComEventCounter
' 0x0C - Get Com Event Log                  ' SGetCommEventLog
' 0x0F - Write Multiple Coils               ' SWriteMultipleCoils
' 0x10 - Write Multiple Reg                 ' SWriteMultipleReg
' 0x11 - Report Slave Id                    ' SReportSlaveId
' 0x16 - Mask Write Register                ' SMaskWriteReg
' 0x17 - Read Write Multiple Regs           ' SReadWriteMultipleRegs
' 0x2B+0x0E - Read Device Identification    ' SReadDeviceIdentification
'--------------------------------------------------------------------------------------
#ifndef Modbus_Protocol_Functions
#define Modbus_Protocol_Functions

#include Modbus_Defs.inc
#include Modbus_CalcCRC_i.inc
#include Modbus_Conf.inc
#include DEFINE_A.INC
#include ufunc4.inc

'' these functions are not implemented yet:
'' 08h Diagnostics
'' 14h ReadFileRecord
'' 15h WriteFileRecord
'' 02Bh/00Dh CANopen General Reference Request and Response PDU



''TEST
String sgTestCoil$
String sgTestDisInput$
String sgTestHoldReg$
String sgTestInputReg$

sub inittest()
    sgTestCoil$ = "<0h>"
    sgTestDisInput$ = "<0h>"
    sgTestHoldReg$ = "<0h><0h>"
    sgTestInputReg$ = "<0h><0h>"
end

''/TEST


WORD	wgEventCounter		'INIT: 0!
WORD	wgProcessedMessages	'INIT: 0!
STRING	sgEvents$(64)		'INIT: <00h>! (contains up to 64 last events (funktion addevent?) => ringbuffer!)
BYTE	bgCurEvPos		'INIT: 0! (contains current position in events string)
BYTE	bgListenOnlyMode	'INIT: 0 = MB_FALSE!
BYTE	bgRunStatus		'INIT: FFh = MB_TRUE = device is working/running


''*************************************************************
''*************************************************************
'' UserData-Functions (read)
''*************************************************************
''*************************************************************

sub GetDeviceIdentifier( byte bpObjectId; var String spObjectData$; var byte bvpReturn)
    bvpReturn = NO_ERROR

    '' Here is the Device-Information specified
    '' the first seven IDs are used for standart device information and should be set
    '' the IDs 80h to FFh are reserved for additional information

    switch bpObjectId
        case 0h:
            spObjectData$="Wilke Technology GmbH"           'VendorName
        case 1h:
            spObjectData$="MB_PROT"                         'ProductCode
        case 2h:
            spObjectData$="V0.0.1"                          'Revision
        case 3h:
            spObjectData$="http://www.wilke.de"             'VendorURL
        case 4h:
            spObjectData$="Modbus protocol for Basic Tiger" 'ProductName
        case 5h:
            spObjectData$="Basic Tiger"                     'ModelName
        case 6h:
            spObjectData$="Modbus_Protocol.inc"             'UserApplicationName

'        case 80h:		'' the addresses 80h to FFh are free for additional
'        case 0FFh:		'' use by the enduser! set Your own information here
        default:
            bvpReturn = NO_SUCH_FUNC_ERROR
            return
    endswitch
end

sub GetUserDataCoils(var String svpUserData$; word wpStartAddress; word wpQuantity; var byte bvpReturn)
    bvpReturn = NO_ERROR

    svpUserData$ = sgTestCoil$

    '' Here the user has to put his COIL data from the given addresses and the given range (wpQuantity)
    '' into the temp-var svpUserData (rigth at the beginning!)
    ''
    '' Addressrange = 0000h to 0FFFFh
    ''
    '' DATA_FORMAT: <---  1. BYTE  ---> <-----   2 .BYTE  ----->...
    '' Values:	    < 0 1 0 0 1 0 0 1 > < 0  0  0  0  1  1 0 1 >...
    '' Adresses:      7 6 5 4 6 2 1 0    15 14 13 12 11 10 9 8  ...
    ''
    '' The bits which were not requestet in the last byte (here 12-15) are to be set to 0!
    '' MAX DATA BYTES = 250! (resulting from a fitting wpQuantity-value, this limit should never be exeeded)
    ''
    '' bvpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Read-Error will be send)

end

sub GetUserDataDisInputs(var string svpUserData$; word wpStartAddress; word wpQuantity; var byte bvpReturn)
    bvpReturn = NO_ERROR	

    svpUserData$ = sgTestDisInput$

    '' Here the user has to put his DISCRETE INPUT data from the given addresses and the given range (wpQuantity)
    '' into the temp-var svpUserData (rigth at the beginning!)
    ''
    '' Addressrange = 0000h to 0FFFFh
    ''
    '' DATA_FORMAT: <---  1. BYTE  ---> <-----   2 .BYTE  ----->...
    '' Values:	    < 0 1 0 0 1 0 0 1 > < 0  0  0  0  1  1 0 1 >...
    '' Adresses:      7 6 5 4 6 2 1 0    15 14 13 12 11 10 9 8  ...
    ''
    '' The bits which were not requestet in the last byte (here 12-15) are to be set to 0!
    '' MAX DATA BYTES = 250! (resulting from a fitting wpQuantity-value, this limit should never be exeeded)
    ''
    '' bvpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Read-Error will be send)

end


sub GetUserDataHoldRegs(var string svpUserData$; word wpStartAddress; word wpQuantity; var byte bvpReturn)
    bvpReturn = NO_ERROR	

    svpUserData$ = sgTestHoldReg$

    '' Here the user has to put his HOLD REGS data from the given addresses and the given range (wpQuantity)
    '' into the temp-var svpUserData (rigth at the beginning!)
    ''
    '' Addressrange = 0000h to 0FFFFh
    ''
    '' DATA_FORMAT:
    ''		<---------- 1. Register  -------------><---------- 2. Register  ------------->...
    ''		<---  Hi BYTE  ---> <---  Lo BYTE  ---><---  Hi BYTE  ---> <---  Lo BYTE  --->...
    '' Data:	< 0 1 0 0 1 0 0 1 > < 0 0 0 0 1 1 0 1 >< 0 1 0 0 1 0 0 1 > < 0 0 0 0 1 1 0 1 >...
    ''
    '' The bits which were not requestet in the last byte (here 12-15) are to be set to 0!
    '' MAX DATA BYTES = 250! (resulting from a fitting wpQuantity-value, this limit should never be exeeded)
    ''
    '' bvpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Read-Error will be send)

end


sub GetUserDataInputRegs(var string svpUserData$; word wpStartAddress; word wpQuantity; var byte bvpReturn)
    bvpReturn = NO_ERROR	

    svpUserData$ = sgTestInputReg$

    '' Here the user has to put his INPUT REGS data from the given addresses and the given range (wpQuantity)
    '' into the temp-var svpUserData (rigth at the beginning!)
    ''
    '' Addressrange = 0000h to 0FFFFh
    ''
    '' DATA_FORMAT:
    ''		<---------- 1. Register  -------------><---------- 2. Register  ------------->...
    ''		<---  Hi BYTE  ---> <---  Lo BYTE  ---><---  Hi BYTE  ---> <---  Lo BYTE  --->...
    '' Data:	< 0 1 0 0 1 0 0 1 > < 0 0 0 0 1 1 0 1 >< 0 1 0 0 1 0 0 1 > < 0 0 0 0 1 1 0 1 >...
    ''
    '' The bits which were not requestet in the last byte (here 12-15) are to be set to 0!
    '' MAX DATA BYTES = 250! (resulting from a fitting wpQuantity-value, this limit should never be exeeded)
    ''
    '' bvpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Read-Error will be send)

end


sub ReadExceptionRegData( var byte bvpData; var byte bvpReturn)
    bvpReturn = NO_ERROR

    bvpData = 0

    '' Here the user has to put his EXCEPTION REGS data
    '' into the temp-var bvpData

end


''*************************************************************
''*************************************************************
'' UserData-Functions (write)
''*************************************************************
''*************************************************************

sub WriteUserDataCoil( word wpAddress; word wpData; var byte bvpReturn)
    bvpReturn = NO_ERROR

    sgTestCoil$ = ""
    sgTestCoil$ = ntos$(sgTestCoil$,0,-SZE_OF_VAR_WORD,wpData)

    '' Here the user has to set the specified coil-address to ON or OFF (depending on wlData)
    '' Addressrange = 0000h to 0FFFFh
    '' wlData = 0000h = OFF; wlData = 0FF00h = ON
    ''
    '' bvpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Write-Error will be send)

end


sub WriteUserDataReg( word wpAddress; word wpData; var byte bvpReturn)
    bvpReturn = NO_ERROR

    sgTestHoldReg$ = ""
    sgTestHoldReg$ = ntos$(sgTestHoldReg$,0,-SZE_OF_VAR_WORD,wpData)

    '' Here the user has to set the specified reg-address to the value of wlData
    '' Addressrange = 0000h to 0FFFFh
    '' wlData-range = 0000h to 0FFFFh
    ''
    '' bpReturn: return-value: 0 = No Error; !=0: error state (MODBUS-Write-Error will be send)

end




''*************************************************************
''*************************************************************
'' Internal ModBus functions:
''*************************************************************
''*************************************************************


''*************************************************************
''**** 0x01 Read Coils
''*************************************************************
Sub SReadCoils( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlQuantity
    word 	wlStartAddress
    byte 	blByteCount
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07D0h then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_COILS, QUANTITY_ERROR)
        return
    endif

    '' Get Start address
    wlStartAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlStartAddress>0FFFFh) or (wlStartAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_COILS, ADDRESS_ERROR)
        return
    endif

    call GetUserDataCoils(slUserData$, wlStartAddress, wlQuantity, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_COILS, READ_ERROR)
        return
    endif

    '' Get the Byte count
    blByteCount=len(slUserData$)

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' Get Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_COILS)

    '' Get quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, blByteCount)


    slMBOutgoingMsg$=slMBOutGoingMsg$+slUserData$

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
end

''*************************************************************
''**** 0x02 Read Discrete Inputs
''*************************************************************
Sub SReadDisInputs( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlQuantity
    word 	wlStartAddress
    byte 	blByteCount
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07D0h then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_DIS_INPUTS, QUANTITY_ERROR)
        return
    endif

    '' Get Start address
    wlStartAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlStartAddress>0FFFFh) or (wlStartAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_DIS_INPUTS, ADDRESS_ERROR)
        return
    endif

    call GetUserDataDisInputs(slUserData$, wlStartAddress, wlQuantity, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_DIS_INPUTS, READ_ERROR)
        return
    endif

    '' Get the Byte count
    blByteCount=len(slUserData$)

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' Get Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DIS_INPUTS)

    '' Get quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, blByteCount)


    slMBOutgoingMsg$=slMBOutGoingMsg$+slUserData$

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End



''*************************************************************
''**** 0x03 Read Hold Reg
''*************************************************************
Sub SReadHoldReg( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlQuantity
    word 	wlStartAddress
    byte 	blByteCount
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07Dh then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_HOLD_REG, QUANTITY_ERROR)
        return
    endif

    '' Get Start address
    wlStartAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlStartAddress>0FFFFh) or (wlStartAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_HOLD_REG, ADDRESS_ERROR)
        return
    endif

    call GetUserDataHoldRegs(slUserData$, wlStartAddress, wlQuantity, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_HOLD_REG, READ_ERROR)
        return
    endif

    '' Get the Byte count
    blByteCount=len(slUserData$)
	
	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' Get Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_HOLD_REG)

    '' Get quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, blByteCount)


    slMBOutgoingMsg$=slMBOutGoingMsg$+slUserData$

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x04 Read Input Reg
''*************************************************************
Sub SReadInputReg( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlQuantity
    word 	wlStartAddress
    byte 	blByteCount
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07Dh then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_INPUT_REG, QUANTITY_ERROR)
        return
    endif

    '' Get Start address
    wlStartAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlStartAddress>0FFFFh) or (wlStartAddress+wlQuantity>0FFFFh) or &
	(wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_INPUT_REG, ADDRESS_ERROR)
    endif

    call GetUserDataInputRegs(slUserData$, wlStartAddress, wlQuantity, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_INPUT_REG, READ_ERROR)
        return
    endif

    '' Get the Byte count
    blByteCount=len(slUserData$)

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' Get Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_INPUT_REG)

    '' Get quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, blByteCount)


    slMBOutgoingMsg$=slMBOutGoingMsg$+slUserData$

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x05 Write Single Coil
''*************************************************************
Sub SWriteSingleCoil( String spMBIncomingMsg$)
    word	wlData
    word 	wlAddress
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    ''Get address
    wlAddress = nfroms(spMBIncomingMsg$,2,-SZE_OF_VAR_WORD)
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_WRITE_SINGLE_COIL, ADDRESS_ERROR)
        return
    endif


    ''Get data
    wlData = nfroms(spMBIncomingMsg$,4,-SZE_OF_VAR_WORD)
    ''ERROR CHECK
    if (wlData <> 0000h) and (wlData <> 0FF00h) then
        '''#ERROR ExceptionCode =03
        call SendError( MBFC_WRITE_SINGLE_COIL, DATA_VALUE_ERROR)
        return
    endif

    call WriteUserDataCoil( wlAddress, wlData, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_WRITE_SINGLE_COIL, WRITE_ERROR)
        return
    endif

	
	'' echo the message back as answer
    call SendMSG( spMBIncomingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif

End


''*************************************************************
''**** 0x06 Write Single Reg
''*************************************************************
Sub SWriteSingleReg( String spMBIncomingMsg$)
    word	wlData
    word 	wlAddress
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    ''Get address
    wlAddress = nfroms(spMBIncomingMsg$,2,-SZE_OF_VAR_WORD)
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_WRITE_SINGLE_REG, ADDRESS_ERROR)
        return
    endif


    ''Get data
    wlData = nfroms(spMBIncomingMsg$,4,-SZE_OF_VAR_WORD)
    ''ERROR CHECK
    if (wlData<0000h) or (wlData>0FFFFh) then
        '''#ERROR ExceptionCode =03
        call SendError( MBFC_WRITE_SINGLE_REG, DATA_VALUE_ERROR)
        return
    endif

    call WriteUserDataReg( wlAddress, wlData, blReturn)
    ''ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_WRITE_SINGLE_REG, WRITE_ERROR)
        return
    endif

	
    '' echo the message back as answer 
    call SendMSG( spMBIncomingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x07 Read Exception Reg
''*************************************************************
Sub SReadExceptionReg( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    word 	wlAddress
    byte	blData
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif


    call ReadExceptionRegData( blData, blReturn)
    ''ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_EXCEPTION_REG, WRITE_ERROR)
        return
    endif


	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_EXCEPTION_REG)

    '' set ExceptionReg data
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DATA_POS_TRANS, SZE_OF_VAR_BYTE, blData)

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x0B Get Com Event Counter
''*************************************************************
Sub SGetComEventCounter( spMBIncomingMsg$)
	String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    byte blReturn

    blReturn = NO_ERROR

    '' assemble modbus response message
    slMBOutGoingMsg$ = ""
    '' set Device ID
 	slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

	'' set function number
	slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_GET_COM_EVENT_COUNTER)

    '' set status (every time 0)
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_STAT_POS_TRANS, -SZE_OF_VAR_WORD, 0)

	'' set event counter
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_EVCT_POS_TRANS, -SZE_OF_VAR_WORD, wgEventCounter)

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
	    wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x0C Get Com Event Log
''*************************************************************
Sub SGetCommEventLog( spMBIncomingMsg$)
    String slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String slEventLog$
    byte blCount
    byte blReturn

    blReturn = NO_ERROR

	
'    call getEventLog(slEventLog$, blReturn)
	
'    blCount = len(slEventLog$) + 6
    blCount = len(sgEvents$) + 6

	
	'' assemble modbus response message
    slMBOutGoingMsg$ = ""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_GET_COM_EVENT_LOG)

    '' set byte count
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DATA_POS_TRANS, SZE_OF_VAR_BYTE, blCount)

    '' set status (every time 0)
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DATA_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD, 0)

    '' set event counter
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_EVCT_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD, wgEventCounter)

    '' set message counter
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MSGC_POS_TRANS+SZE_OF_VAR_BYTE, -SZE_OF_VAR_WORD, wgProcessedMessages)

    '' add Eventlog
    slMBOutGoingMsg$=slMBOutGoingMsg$ + sgEvents$


    call SendMSG( slMBOutGoingMsg$, blReturn)
    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** get event log
''*************************************************************
sub getEventLog( var String spEventLog$; var byte spReturn)
    spReturn = 0

    spEventLog$ = ""

    if len(sgEvents$) < bgCurEvPos then
        spEventLog$ = stos$(spEventLog$, 0, sgEvents$, -bgCurPos-1, len(sgEvents$)-bgCurEvPos)
        spEventLog$ = stos$(spEventLog$, bgCurEvPos, sgEvents$, 0, bgCurEvPos-1)
    else
        spEventLog$ = sgEvents$
    endif

end


''*************************************************************
''**** 0x0F Write Multiple Coils
''*************************************************************
Sub SWriteMultipleCoils( String spMBIncomingMsg$)
    String	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    word	wlData
    word 	wlAddress
    byte	blByteCount
    byte 	blReturn
    byte 	blCurPos

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif


    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07D0h then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_WRITE_MULT_COILS, QUANTITY_ERROR)
        return
    endif

    ''Get address
    wlAddress = nfroms(spMBIncomingMsg$,MB_ADDR_POS_TRANS,-SZE_OF_VAR_WORD)
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlAddress+wlQuantity>0FFFFh) or &
            (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_WRITE_MULT_COILS, ADDRESS_ERROR)
        return
    endif


    '' Get the byte count
    blByteCount=nfroms(spMBIncomingMsg$, 6,-SZE_OF_VAR_BYTE)
    if wlQuantity shr 3 <> blByteCount AND wlQuantity shr 3 + 1 <> blByteCount then
        call SendError( MBFC_WRITE_MULT_COILS, QUANTITY_ERROR)
        return
    endif

    blCurBit = 0
    blCurByte = 7

    for blByteCount=wlQuantity to 1 step -1
        ''Get data
        if bit(nfroms(spMBIncomingMsg$,blCurByte,-SZE_OF_VAR_BYTE),blCurBit) = 0 then
            call WriteUserDataCoil( wlAddress, 0000H, blReturn)
        else
            call WriteUserDataCoil( wlAddress, 0FF00H, blReturn)
        endif

        '' ERROR CHECK
        if blReturn <> NO_ERROR then
            '''# ERROR ExceptionCode = 04h
            call SendError( MBFC_WRITE_MULT_COILS, WRITE_ERROR)
            return
        endif


        wlAddress = wlAddress+1
        if blCurBit = 7 then
            blCurBit = 0
            blCurByte = blCurByte+1
        else
            blCurBit = blCurBit+1
        endif
    next


	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_MULT_COILS)

    wlAddress = nfroms(spMBIncomingMsg$,MB_ADDR_POS_TRANS,-SZE_OF_VAR_WORD)
    '' set start address
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wlAddress)

    '' set quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DATA_POS_TRANS+SZE_OF_VAR_WORD, -SZE_OF_VAR_WORD, wlQuantity)


    call SendMSG( slMBOutGoingMsg$, blReturn)
    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif

End



''*************************************************************
''**** 0x10 Write Multiple Reg
''*************************************************************
Sub SWriteMultipleReg( String spMBIncomingMsg$)
    String	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    word	wlData
    word 	wlAddress
    byte	blCurByte
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 007Bh then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_WRITE_MULT_REG, QUANTITY_ERROR)
        return
    endif

    ''Get address
    wlAddress = nfroms(spMBIncomingMsg$,2,-SZE_OF_VAR_WORD)
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlAddress+wlQuantity>0FFFFh) or &
            (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_WRITE_MULT_REG, ADDRESS_ERROR)
        return
    endif


    '' Get the byte count
    blByteCount=nfroms(spMBIncomingMsg$, 6,-SZE_OF_VAR_BYTE)		
    '' ERROR CHECK
    if blByteCount <> wlQuantity*2 then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_WRITE_MULT_REG, QUANTITY_ERROR)
        return
    endif


    blCurByte = 7

    for blByteCount=blByteCount - 2 to 0 step -2
        ''Get data
        wlData = nfroms(spMBIncomingMsg$,blCurByte,-SZE_OF_VAR_BYTE)
        ''ERROR CHECK
        if (wlData<0000h) or (wlData>0FFFFh) then
            '''#ERROR ExceptionCode =03
            call SendError( MBFC_WRITE_MULT_REG, DATA_VALUE_ERROR)
            return
        endif

        call WriteUserDataReg( wlAddress, wlData, blReturn)
        ''ERROR CHECK
        if blReturn <> NO_ERROR then
            '''# ERROR ExceptionCode = 04h
            call SendError( MBFC_WRITE_MULT_REG, WRITE_ERROR)
            return
        endif

        wlAddress = wlAddress+1
        blCurByte = wlCurByte+2
    next

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_WRITE_MULT_REG)

    wlAddress = nfroms(spMBIncomingMsg$,MB_ADDR_POS_TRANS,-SZE_OF_VAR_WORD)
    '' set start address
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ADDR_POS_TRANS, -SZE_OF_VAR_WORD, wlAddress)

    '' set quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DATA_POS_TRANS+SZE_OF_VAR_WORD, -SZE_OF_VAR_WORD, wlQuantity)


    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x11 Report Slave Id
''*************************************************************
Sub SReportSlaveId( String spMBIncomingMsg$)
    String	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slDeviceId$
    byte	blReturn

    blReturn = NO_ERROR


    slDeviceId$=DEVICE_ID

    slDeviceId$=ntos$(slDeviceId$,len(slDeviceId$), SZE_OF_VAR_BYTE, bgRunStatus)

    slDeviceId$=slDeviceId$+ADDITIONAL_DEVICE_INFORMATION

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_REPORT_SLAVE_ID)

    '' set byte count
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, len(DEVICE_ID))

    slMBOutGoingMsg$=slMBOutGoingMsg$+slDeviceId$

    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** 0x16 Mask Write Register
''*************************************************************
Sub SMaskWriteReg( String spMBIncomingMsg$)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlAddress
    word 	wlData
    word	wlAndMask
    word	wlOrMask
    word	wlNotAndMask
    byte 	blReturn

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif

    '' Get Start address
    wlAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_MASK_WRITE_REGISTER, ADDRESS_ERROR)
        return
    endif


    call GetUserDataHoldRegs(slUserData$, wlAddress, SZE_OF_VAR_BYTE, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_MASK_WRITE_REGISTER, READ_ERROR)
        return
    endif

    wlNotAndMask=BITNOT(wlAndMask)
    wlData=nfroms(slUserData$,0,-SZE_OF_VAR_WORD)
    wlData=(wlData BITAND wlAndMask)BITOR(wlOrMask BITAND wlNotAndMask)


    call WriteUserDataReg( wlAddress, wlData, blReturn)
    ''ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_MASK_WRITE_REGISTER, WRITE_ERROR)
        return
    endif

	'' echo the message back as answer
    call SendMSG( spMBIncomingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End



''*************************************************************
''**** 0x17 Read Write Multiple Regs
''*************************************************************
Sub SReadWriteMultipleRegs( String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slUserData$(USER_DATA_MAX_STR_SZE)
    word 	wlQuantity
    word 	wlStartAddress
    byte 	blByteCount
    byte 	blReturn
    word	wlData
    word 	wlAddress
    byte	blCurByte

    blReturn = NO_ERROR


    if spMBIncomingMsg$="" then
        return
    endif


    ''write the regs

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 8,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 007Bh then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_WRITE_MULT_REGS, QUANTITY_ERROR)
        return
    endif

    ''Get address
    wlAddress = nfroms(spMBIncomingMsg$,6,-SZE_OF_VAR_WORD)
    '' ERROR CHECK
    if (wlAddress>0FFFFh) or (wlAddress+wlQuantity>0FFFFh) or &
            (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_WRITE_MULT_REGS, ADDRESS_ERROR)
        return
    endif


    '' Get the byte count
    blByteCount=nfroms(spMBIncomingMsg$, 10,-SZE_OF_VAR_BYTE)		
    '' ERROR CHECK
    if blByteCount <> wlQuantity*2 then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_WRITE_MULT_REGS, QUANTITY_ERROR)
        return
    endif


    blCurByte = 11

    for blByteCount=blByteCount -2 to 0 step -2
        ''Get data
        wlData = nfroms(spMBIncomingMsg$,blCurByte,-SZE_OF_VAR_WORD)
        ''ERROR CHECK
        if (wlData<0000h) or (wlData>0FFFFh) then
            '''#ERROR ExceptionCode =03
            call SendError( MBFC_READ_WRITE_MULT_REGS, DATA_VALUE_ERROR)
            return
        endif

        call WriteUserDataReg( wlAddress, wlData, blReturn)
        ''ERROR CHECK
        if blReturn <> NO_ERROR then
            '''# ERROR ExceptionCode = 04h
            call SendError( MBFC_READ_WRITE_MULT_REGS, WRITE_ERROR)
            return
        endif

        wlAddress = wlAddress+1
        blCurByte = blCurByte+2
    next

    '' read the regs

    '' Get the Quantity
    wlQuantity=nfroms(spMBIncomingMsg$, 4,-SZE_OF_VAR_WORD)		
    '' ERROR CHECK
    if wlQuantity<0000h or wlQuantity > 07Dh then
        '''# ERROR ExceptionCode = 03h
        call SendError( MBFC_READ_WRITE_MULT_REGS, QUANTITY_ERROR)
        return
    endif

    '' Get Start address
    wlStartAddress=nfroms(spMBIncomingMsg$, 2,-SZE_OF_VAR_WORD)	
    '' ERROR CHECK
    if (wlStartAddress>0FFFFh) or (wlStartAddress+wlQuantity>0FFFFh) or &
        (wlStartAddress<0000h) then
        '''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_WRITE_MULT_REGS, ADDRESS_ERROR)
        return
    endif

    call GetUserDataHoldRegs(slUserData$, wlStartAddress, wlQuantity, blReturn)
    '' ERROR CHECK
    if blReturn <> NO_ERROR then
        '''# ERROR ExceptionCode = 04h
        call SendError( MBFC_READ_WRITE_MULT_REGS, READ_ERROR)
        return
    endif

    '' Get the Byte count
    blByteCount=len(slUserData$)

	
	'' assemble modbus response message
    slMBOutGoingMsg$=""
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' Get Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_WRITE_MULT_REGS)

    '' Get quantity
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_QATY_POS_TRANS, SZE_OF_VAR_BYTE, blByteCount)

    slMBOutgoingMsg$=slMBOutGoingMsg$+slUserData$


    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End



''*************************************************************
''**** 0x02B+0x0E Read Device Identification
''*************************************************************
Sub SReadDeviceIdentification( String spMBIncomingMsg$)
    byte	blReadDevIdCode
    byte 	blByteCount
    byte	blReqObjectId
    byte	blNumberOfObjects
    String	slObjectData$


    blReadDevIdCode=nfroms(spMBIncomingMsg$,3,SZE_OF_VAR_BYTE)
    ''ERROR CHECK
    if (blReadDevIdCode<1h) or (blReadDevIdCode>04h) then
        '''#ERROR ExceptionCode =03
        call SendError( MBFC_READ_DEV_IDENTIFICATION, DATA_VALUE_ERROR)
        return
    endif

    blReqObjectId=nfroms(spMBIncomingMsg$,4,SZE_OF_VAR_BYTE)
    '' ERROR CHECK
'    if ((blReqObjectId<0h) or (blReqObjectId>06h)) AND&
'      ((blReqObjectId<80h) or (blReqObjectId>0FFh)) then		'not supported now
    if ((blReqObjectId<0h) or (blReqObjectId>06h)) then
	'''# ERROR ExceptionCode = 02h
        call SendError( MBFC_READ_DEV_IDENTIFICATION, ADDRESS_ERROR)
        return
    endif

    switch (blReadDevIdCode)
        case 1:
            call SendBasicDeviceIdent(spMBIncomingMsg$)

        case 2:
            call SendRegularDeviceIdent(spMBIncomingMsg$)

        case 3:
            ''Not supported => address_error
            call SendError( MBFC_READ_DEV_IDENTIFICATION, ADDRESS_ERROR)

        case 4:
            call SendSpecificDeviceIdent(blReqObjectId)

    endswitch

End


''*************************************************************
''**** send the basic device identifiers (subfunction for 0x02Bh)
''*************************************************************
Sub SendBasicDeviceIdent(String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slObjectData$(MAX_IDENT_OBJ_DATA_SZE) '245
    String	slTempString$(MAX_IDENT_OBJ_DATA_SZE) '245
    byte	blCount
    byte 	blReturn
    byte	blFollows
    byte	blCounter
    byte	blObjectId

    blReturn = NO_ERROR
    slMBOutGoingMsg$=""
    slTempString$=""
    blCounter = 0

    blObjectId = nfroms(spMBIncomingMsg$,4,SZE_OF_VAR_BYTE)

	
    '' assemble modbus response message
    for blCount = blObjectId to 2
        '' if we have to build a new response (string is empty):
        if len(slMBOutGoingMsg$)=0 then
            '' set Device ID
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

            '' set Function number
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_IDENTIFICATION)

            '' set MEI type
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ADDR_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_MEI_TYPE)

            '' set Read Device Id Code
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DVID_POS_TRANS, SZE_OF_VAR_BYTE, 81h)

            '' set conformity level
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_BASIC_FULL_ACC)
        endif

        '' get data
        call GetDeviceIdentifier( blCount, slObjectData$, blReturn)
        if blReturn <> NO_ERROR then
            call SendError( MBFC_READ_DEV_IDENTIFICATION, ADDRESS_ERROR)
        endif

        blFollows = 0FFh

        '' if there is enough space left for the next object, add it
        if (len(slMBOutGoingMsg$)+len(slTempString$)+len(slObjectData$)+5)<=OUT_GOING_STR_MAX_SZE then
            '' object id
            slTempString$=ntos$(slTempString$, len(slTempString$), SZE_OF_VAR_BYTE, blCount)

            '' object length
            slTempString$=ntos$(slTempString$, len(slTempString$), SZE_OF_VAR_BYTE, len(slObjectData$))

            slTempString$=slTempString$+slObjectData$

            blFollows = 0
            blCounter = blCounter + 1
        endif

        '' are there any objects left? => build a fitting response
        if blFollows = 0FFh then
            '' more follows
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE, 0FFh)

            '' next object
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE, blCount+1)

            '' number of objects
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_OBJC_POS_TRANS, SZE_OF_VAR_BYTE, blCount+1)

            slMBOutGoingMsg$=slMBOutGoingMsg$+slTempString$

            call SendMSG( slMBOutGoingMsg$, blReturn)

            return

        else
            '' if the last object is in this package:
            if blCount = 2 then
                '' no more follows
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE, 0h)

                '' no next object
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE, 0h)

                '' number of objects
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_OBJC_POS_TRANS, SZE_OF_VAR_BYTE, blCounter)

                slMBOutGoingMsg$=slMBOutGoingMsg$+slTempString$

                call SendMSG( slMBOutGoingMsg$, blReturn)

                return
            endif
        endif

    next

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif

End


''*************************************************************
''**** send the regular device identifiers (subfunction for 0x02Bh)
''*************************************************************
Sub SendRegularDeviceIdent(String spMBIncomingMsg$)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slObjectData$(MAX_IDENT_OBJ_DATA_SZE) '245
    String	slTempString$(MAX_IDENT_OBJ_DATA_SZE) '245
    byte	blCount
    byte 	blReturn
    byte	blFollows
    byte	blCounter
    byte	blObjectId

    blReturn = NO_ERROR
    slMBOutGoingMsg$=""
    slTempString$=""
    blCounter = 0

    blObjectId = nfroms(spMBIncomingMsg$,4,SZE_OF_VAR_BYTE)

    if blObjectId < 3 then
        blObjectId = 3
    endif

	
    '' assemble modbus response message
    for blCount = blObjectId to 6
	    '' if we have to build a new response (string is empty):
        if len(slMBOutGoingMsg$)=0 then
            '' set Device ID
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

            '' set Function number
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_IDENTIFICATION)

            '' set MEI type
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ADDR_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_MEI_TYPE)

            '' set Read Device Id Code
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DVID_POS_TRANS, SZE_OF_VAR_BYTE, 82h)

            '' set conformity level
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_REGULAR_FULL_ACC)
        endif

        '' get data
        call GetDeviceIdentifier( blCount, slObjectData$, blReturn)
        if blReturn <> NO_ERROR then
            call SendError( MBFC_READ_DEV_IDENTIFICATION, ADDRESS_ERROR)
        endif

        blFollows = 0FFh

        '' if there is enough space left for the next object, add it
        if (len(slMBOutGoingMsg$)+len(slTempString$)+len(slObjectData$)+5)<=OUT_GOING_STR_MAX_SZE then
            '' object id
            slTempString$=ntos$(slTempString$, len(slTempString$), SZE_OF_VAR_BYTE, blCount)

            '' object length
            slTempString$=ntos$(slTempString$, len(slTempString$), SZE_OF_VAR_BYTE, len(slObjectData$))

            slTempString$=slTempString$+slObjectData$

            blFollows = 0
            blCounter = blCounter + 1
        endif

        '' are there any objects left? => build a fitting response
        if blFollows = 0FFh then
            '' more follows
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE, 0FFh)

            '' next object
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE, blCount+1)

            '' number of objects
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_OBJC_POS_TRANS, SZE_OF_VAR_BYTE, blCounter)

            slMBOutGoingMsg$=slMBOutGoingMsg$+slTempString$

            call SendMSG( slMBOutGoingMsg$, blReturn)

            return

        else
            '' if the last object is in this package:
            if blCount = 6 then
                '' no more follows
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE, 0h)

                '' no next object
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE, 0h)

                '' number of objects
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_OBJC_POS_TRANS, SZE_OF_VAR_BYTE, blCounter)

                slMBOutGoingMsg$=slMBOutGoingMsg$+slTempString$

                call SendMSG( slMBOutGoingMsg$, blReturn)

                return
            endif
        endif

    next

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif

End


''*************************************************************
''**** send a specific device identifier (subfunction for 0x02Bh)
''*************************************************************
Sub SendSpecificDeviceIdent( byte bpReqObjectId)
    String 	slMBOutGoingMsg$(OUT_GOING_STR_MAX_SZE)
    String	slObjectData$(MAX_IDENT_OBJ_DATA_SZE) '245
    byte 	blReturn

    blReturn = NO_ERROR

    '' get data
    call GetDeviceIdentifier( bpReqObjectId, slObjectData$, blReturn)
    if blReturn <> NO_ERROR then
        call SendError( MBFC_READ_DEV_IDENTIFICATION, ADDRESS_ERROR)
    endif

	
    '' assemble modbus response message
    slMBOutGoingMsg$=""
	
    '' set Device ID
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE, MY_DEVICE_ID)

    '' set Function number
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_IDENTIFICATION)

    '' set MEI type
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ADDR_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_READ_DEV_MEI_TYPE)

    '' set conformity level
	if bpReqObjectId <= 2 then
        slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_BASIC_FULL_ACC)
    else
        if bpReqObjectId <= 6 then
            slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_REGULAR_FULL_ACC)
        else
            if bpReqObjectId >= 080h then
                slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_COLV_POS_TRANS, SZE_OF_VAR_BYTE, MBFC_EXTENDED_ACC)
            endif
        endif
    endif

    '' no more follows
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_MFLW_POS_TRANS, SZE_OF_VAR_BYTE, 0)

    '' no next object
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_NOBJ_POS_TRANS, SZE_OF_VAR_BYTE, 0)

    '' number of objects
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_OBJC_POS_TRANS, SZE_OF_VAR_BYTE, 1)

    '' object id
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ROID_POS_TRANS, SZE_OF_VAR_BYTE, bpReqObjectId)

    '' object length
    slMBOutGoingMsg$=ntos$(slMBOutGoingMsg$, MB_ROLE_POS_TRANS, SZE_OF_VAR_BYTE, len(slObjectData$))

    '' append object data
    slMBOutGoingMsg$=slMBOutGoingMsg$+slObjectData$

    '' send
    call SendMSG( slMBOutGoingMsg$, blReturn)

    if blReturn = NO_ERROR then
        wgEventCounter = wgEventCounter + 1
    endif
End


''*************************************************************
''**** add event - adds an event to the eventcounter
''*************************************************************
sub addEvent( long lpEvent; long lpEventParam; long lpErrorCode; var byte bpReturn)
    bpReturn = 0

    byte blLogEntry
    blLogEntry = 0

    if bgCurEvPos = max_len(sgEvents$)-1 then
        bgCurEvPos = 0
    else
        bgCurEvPos = bgCurEvPos + 1
    endif


    if lpEvent = MB_RECEIVE_EVENT then
	blLogEntry = blLogEntry BITOR 80h

        if lpEventParam = MB_BROADCAST_MSG then
            blLogEntry = blLogEntry BITOR 40h
        endif

        if bgListenOnlyMode = MB_TRUE then
            blLogEntry = blLogEntry BITOR 20h		
        endif

        if bpErrorCode <> NO_ERROR then
            blLogEntry = blLogEntry BITOR 2h
        endif

	sgEvents$ = ntos$(sgEvents$,bgCurEvPos,SZE_OF_VAR_BYTE,blLogEntry)
        return
    endif

    if lpEvent = MB_SEND_EVENT then
        blLogEntry = blLogEntry BITOR 40h

        switchi lpErrorCode

        case NO_SUCH_FUNC_ERROR:
            blLogEntry = blLogEntry BITOR 1h

        case ADDRESS_ERROR:
            blLogEntry = blLogEntry BITOR 1h

        case DATA_VALUE_ERROR:
            blLogEntry = blLogEntry BITOR 1h

        case READ_ERROR:	'also write_error
            blLogEntry = blLogEntry BITOR 2h

        case 5h:
            blLogEntry = blLogEntry BITOR 4h

        case 6h:
            blLogEntry = blLogEntry BITOR 4h

        case 7h:
            blLogEntry = blLogEntry BITOR 8h

        endswitch

        if bgListenOnlyMode = MB_TRUE then
            blLogEntry = blLogEntry BITOR 20h		
        endif

	sgEvents$ = ntos$(sgEvents$,bgCurEvPos,SZE_OF_VAR_BYTE,blLogEntry)
        return
    endif

    if lpEvent = MB_ENTER_LISTEN_ONLY then
        sgEvents$ = stos$(sgEvents$,bgCurEvPos,"<04h>",0,1)
        return
    endif

    if lpEvent = MB_RESTART_COMM then
        sgEvents$ = stos$(sgEvents$,bgCurEvPos,"<00h>",0,1)
        return
    endif

end


''*************************************************************
''**** Send Error - Sends an MB-Error for the specified function and error
''*************************************************************
sub SendError ( byte bpFunctionCode; byte bpErrorCode)
	string slErrorMSG$(2)
        byte blReturn

	slErrorMSG$=""

	'' Get funktion code
        slErrorMSG$=ntos$( slErrorMSG$,0, SZE_OF_VAR_BYTE, MY_DEVICE_ID)
	slErrorMSG$=ntos$( slErrorMSG$,1, SZE_OF_VAR_BYTE, bpFunctionCode+80h)
	slErrorMSG$=ntos$( slErrorMSG$,2, SZE_OF_VAR_BYTE, bpErrorCode)

        call SendMSG(slErrorMSG$, blReturn)
end


''*************************************************************
''**** Send Message - Sends the MB-Message to other devices
''*************************************************************
sub SendMSG( String spMBMessage$; var byte bpReturn)
    String slMessage$(OUT_GOING_STR_MAX_SZE)
    word wlCRC
    byte blEvent
    byte blEventParam
    byte blErrorCode

    wlCRC = 0
    bpReturn = 0
    slMessage$ = spMBMessage$

    if bgListenOnlyMode = MB_FALSE then
        call calcCRC( slMessage$, wlCRC)
        slMessage$ = ntos$(slMessage$,len(slMessage$),SZE_OF_VAR_WORD,wlCRC)

		wait_duration 10
		PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0
        PUT #SER, slMessage$
    endif

    call addEvent (MB_SEND_EVENT, 0, 0, bpReturn)
end


''*************************************************************
''**** Handle a Modbus Message
''*************************************************************
sub HandleMBMessage( String spMBMsg$)
	long llAdrInMsg
	byte blFuncInMsg
        byte blReturn
	
	llAdrInMsg= nfroms(spMbMsg$, MB_DEVC_POS_TRANS, SZE_OF_VAR_BYTE)

	if llAdrInMsg = MY_DEVICE_ID or MB_BROADCAST_MSG then
		blFuncInMsg=nfroms(spMbMsg$, MB_FUNC_POS_TRANS, SZE_OF_VAR_BYTE)
		'' ModBus Msg include the function code
		switch blFuncInMsg
		case 01h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadCoils(spMBMsg$)
		case 02h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadDisInputs(spMBMsg$)
		case 03h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadHoldReg(spMBMsg$)
		case 04h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadInputReg(spMBMsg$)
		case 05h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SWriteSingleCoil(spMBMsg$)
		case 06h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SWriteSingleReg(spMBMsg$)
		case 07h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadExceptionReg(spMBMsg$)
		case 0Bh:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SGetComEventCounter(spMBMsg$)
		case 0Ch:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
                        call SGetCommEventLog(spMBMsg$)
		case 0Fh:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SWriteMultipleCoils(spMBMsg$)
		case 010h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SWriteMultipleReg(spMBMsg$)
		case 011h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReportSlaveId(spMBMsg$)
		case 016h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
                        call SMaskWriteReg(spMBMsg$)
		case 017h:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
			call SReadWriteMultipleRegs(spMBMsg$)
		case 02Bh:
                        if nfroms(spMBMsg$,MB_QATY_POS_TRANS,SZE_OF_VAR_BYTE)=0EH then
                            call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_ERROR, blReturn)
		            call SReadDeviceIdentification(spMBMsg$)
                        else
                            call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_SUCH_FUNC_ERROR, blReturn)
                            call SendError(blFuncInMsg, NO_SUCH_FUNC_ERROR)
                        endif
		default:
                        call addEvent (MB_RECEIVE_EVENT, llAdrInMsg, NO_SUCH_FUNC_ERROR, blReturn)
			call SendError(blFuncInMsg, NO_SUCH_FUNC_ERROR)
		endswitch	
	endif
        PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0        	
end

task ModbusProtocol
    word wlIBUFill
    word wlMsgCRC
    word wlCRC
    String slMBMessage$(255)

    wlCRC = 0
    wlMsgCRC = 0

    PUT #SER, #SEC_RS485, #UFCO_IBU_ERASE,0

    while 1=1
        GET #SER, #SEC_RS485, #UFCI_IBU_FILL, 0, wlIBUFill
        if wlIBUFill <> 0 then
            wait_duration 5
            GET #SER, #SEC_RS485, 0, slMBMessage$

            if len(slMBMessage$) >= 4 then
                wlMsgCRC = nfroms(slMBMessage$, len(slMBMessage$)-2, SZE_OF_VAR_WORD)
                slMBMessage$ = left$(slMBMessage$, len(slMBMessage$)-2)
                call calcCRC(slMBMessage$,wlCRC)
                wgProcessedMessages = wgProcessedMessages + 1
                if wlCRC = wlMsgCRC then
                    call HandleMBMessage(slMBMessage$)
                endif
            endif
        else
            wait_duration 10
        endif
    endwhile
end



''*************************************************************
''**** Modbus Init
''*************************************************************
Sub InitModbus()
    wgEventCounter = 0          'INIT: 0!
    wgProcessedMessages = 0     'INIT: 0!
    sgEvents$ = "<00h>"         'INIT: <00h>! (contains up to 64 last events => ringbuffer)
    bgCurEvPos = 0              'INIT: 0! (contains current position in events string)
    bgListenOnlyMode = MB_FALSE 'INIT: 0 = MB_FALSE!
    bgRunStatus = MB_TRUE       'INIT: FFh = MB_TRUE = device is working/running
    call CRCinit()
    INSTALL_DEVICE #SER, "SER1B_K4.TDD",&
        BAUD_SEC_0, PARITY_SEC_0, REC_FLAG_SEC_0,&
        BAUD_SEC_1, PARITY_SEC_1, REC_FLAG_SEC_1,&
        TR_EN_MASK_SEC_0,LOG_PORT_SEC_0,TR_HILO_SEC_0,&
        TR_EN_MASK_SEC_1,LOG_PORT_SEC_1,TR_HILO_SEC_1
    Run_Task ModbusProtocol
End
#endif

